from pathlib import Path

from kaizo import ConfigParser

X = 5

module_config = f"""
x: {X}
"""

config = """
import:
  m: {tmp_path}/module.yml
run: m.x
"""


def test_import_module(tmp_path: Path) -> None:
    module = tmp_path / "module.yml"
    module.write_text(module_config)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(config.format(tmp_path=tmp_path))

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    assert out["run"] == X



from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable, MutableMapping, MutableSequence
from dataclasses import dataclass
from typing import Any, Generic, SupportsIndex, TypeVar

from .fn import FnWithKwargs

K = TypeVar("K")
V = TypeVar("V")


@dataclass
class Entry(ABC):
    key: str

    @abstractmethod
    def __call__(self) -> Any:
        pass


class DictEntry(MutableMapping, Generic[K]):
    _data: dict[K, Entry]
    _resolve: bool

    def __init__(self, data: dict[K, Entry] | None = None, *, resolve: bool = True) -> None:
        super().__init__()

        self._data = data or {}
        self._resolve = resolve

    def __setitem__(self, key: K, value: Any) -> None:
        if not isinstance(value, Entry):
            msg = f"Value must be an Entry instance, got {type(value)}"
            raise TypeError(msg)

        self._data[key] = value

    def __getitem__(self, key: K) -> Any:
        value = self._data[key]

        if not isinstance(value, Entry):
            msg = f"Value must be an Entry instance, got {type(value)}"
            raise TypeError(msg)

        if self._resolve:
            return value.__call__()

        return value

    def __delitem__(self, key: K) -> None:
        self._data.__delitem__(key)

    def __iter__(self) -> Generator[K]:
        yield from self._data

    def __len__(self) -> int:
        return self._data.__len__()

    def __contains__(self, key: K) -> bool:
        return self._data.__contains__(key)


class ListEntry(MutableSequence):
    _data: list[Entry]
    _resolve: bool

    def __init__(self, data: list[Entry] | None = None, *, resolve: bool = True) -> None:
        super().__init__()

        self._data = data or []
        self._resolve = resolve

    def __setitem__(self, i: SupportsIndex, value: Any) -> None:
        if isinstance(value, Iterable):
            for value_i in value:
                if not isinstance(value_i, Entry):
                    msg = f"Value must be an Entry instance, got {type(value_i)}"
                    raise TypeError(msg)

        elif not isinstance(value, Entry):
            msg = f"Value must be an Entry instance, got {type(value)}"
            raise TypeError(msg)

        self._data[i] = value

    def __getitem__(self, i: SupportsIndex) -> Any:
        value = self._data.__getitem__(i)

        if isinstance(value, Iterable):
            new_values = []
            for value_i in value:
                if not isinstance(value_i, Entry):
                    msg = f"Value must be an Entry instance, got {type(value_i)}"
                    raise TypeError(msg)

                if self._resolve:
                    new_values.append(value_i.__call__())
                else:
                    new_values.append(value_i)

            return new_values

        if isinstance(value, Entry):
            if self._resolve:
                return value.__call__()

            return value

        msg = f"Value must be an Entry instance, got {type(value)}"
        raise TypeError(msg)

    def __delitem__(self, i: SupportsIndex) -> None:
        self._data.__delitem__(i)

    def __len__(self) -> int:
        return self._data.__len__()

    def insert(self, index: SupportsIndex, value: Any) -> None:
        if not isinstance(value, Entry):
            msg = f"Value must be an Entry instance, got {type(value)}"
            raise TypeError(msg)

        self._data.insert(index, value)


@dataclass
class FieldEntry(Entry, Generic[V]):
    value: V

    def __call__(self) -> V:
        return self.value


@dataclass
class ModuleEntry(Entry):
    obj: Any
    call: Any
    lazy: bool
    args: DictEntry[str] | ListEntry | None = None
    cache: Any | None = None

    def _cache_call(self, fn: Any, args: tuple, kwargs: dict) -> Any:
        if self.cache is None:
            if self.lazy:
                self.cache = FnWithKwargs(fn=fn, args=args, kwargs=kwargs)
            else:
                self.cache = fn(*args, **kwargs)

        return self.cache

    def __call__(self) -> Any | FnWithKwargs:
        kwargs = {}
        args = ()

        if isinstance(self.args, DictEntry):
            kwargs = self.args
        elif isinstance(self.args, ListEntry):
            args = self.args

        if self.call is False:
            return self.obj

        if self.call is True:
            if not callable(self.obj):
                msg = f"'{self.obj}' is not callable"
                raise TypeError(msg)

            return self._cache_call(self.obj, args, kwargs)

        if not hasattr(self.obj, self.call):
            msg = f"'{self.obj}' has no attribute '{self.call}'"
            raise AttributeError(msg)

        fn = getattr(self.obj, self.call)

        if not callable(fn):
            msg = f"{fn} is not callable"
            raise TypeError(msg)

        return self._cache_call(fn, args, kwargs)



from .common import split_by_first_dot
from .entry import DictEntry, Entry, FieldEntry, ListEntry, ModuleEntry
from .fn import FnWithKwargs

__all__ = (
    "DictEntry",
    "Entry",
    "FieldEntry",
    "FnWithKwargs",
    "ListEntry",
    "ModuleEntry",
    "split_by_first_dot",
)



from pathlib import Path

from kaizo import ConfigParser
from kaizo.utils import FnWithKwargs

VAL = 9

config = f"""
fn:
  module: math
  source: sqrt
  call: true
  lazy: true
  args:
    - {VAL}
"""


def test_lazy_call(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    entry = out["fn"]
    assert isinstance(entry, FnWithKwargs)

    assert entry() == VAL**0.5



import importlib
import os
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from types import ModuleType
from typing import Any

import yaml

from .utils import DictEntry, Entry, FieldEntry, ListEntry, ModuleEntry, split_by_first_dot


class ConfigParser:
    config: dict[str]
    local: ModuleType | None
    variables: DictEntry[str]
    kwargs: dict[str]
    modules: dict[str, "ConfigParser"] | None

    def __init__(self, config_path: str | Path, kwargs: dict[str] | None = None) -> None:
        root, _ = os.path.split(config_path)

        root = Path(root)

        with Path.open(config_path) as file:
            self.config = yaml.safe_load(file)

        if "local" in self.config:
            local_path = Path(self.config.pop("local"))
            self.local = self._load_python_module(root / local_path)
        else:
            self.local = None

        if "import" in self.config:
            modules = self.config.pop("import")

            if not isinstance(modules, dict):
                msg = f"import module should be a dict, got {type(modules)}"
                raise TypeError(msg)

            self.modules = self._import_modules(modules, kwargs)
        else:
            self.modules = None

        self.variables = DictEntry(resolve=False)
        self.kwargs = kwargs or {}

    def _load_python_module(self, path: Path) -> ModuleType:
        if not path.is_file():
            msg = f"Local Python file not found: {path}"
            raise FileNotFoundError(msg)

        module_name = path.stem
        spec = spec_from_file_location(module_name, path)
        if spec is None or spec.loader is None:
            msg = f"Failed to load module from: {path}"
            raise ImportError(msg)

        module = module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def _import_modules(
        self,
        modules: dict[str, str],
        kwargs: dict[str] | None = None,
    ) -> dict[str, "ConfigParser"]:
        module_dict = {}

        for module_name, module_path in modules.items():
            parser = ConfigParser(module_path, kwargs)
            parser.parse()

            module_dict[module_name] = parser

        return module_dict

    def _load_object(self, module_path: str, object_name: str) -> Any:
        try:
            module = importlib.import_module(module_path)
            if not hasattr(module, object_name):
                msg = f"Module '{module_path}' has no attribute '{object_name}'"
                raise AttributeError(msg)
            return getattr(module, object_name)
        except ModuleNotFoundError as e:
            msg = f"Could not import module '{module_path}': {e}"
            raise ImportError(msg) from e

    def _load_symbol_from_module(self, module_path: str, symbol_name: str) -> Any:
        if module_path == "local":
            if self.local is None:
                msg = "local module is not given"
                raise ValueError(msg)

            return getattr(self.local, symbol_name)

        return self._load_object(module_path, symbol_name)

    def _resolve_string(self, key: str, entry: str) -> Entry:
        entry_key, entry_value = split_by_first_dot(entry)

        if entry_key is None:
            return FieldEntry(key=key, value=entry)

        if not entry_key:
            return (
                FieldEntry(key=entry_value, value=self.kwargs[entry_value])
                if entry_value in self.kwargs
                else self.variables.get(entry_value)
            )

        if self.modules is None:
            msg = "import module is not given"
            raise ValueError(msg)

        module = self.modules.get(entry_key)

        if module is None:
            msg = f"keyword not found, got {entry_key}"
            raise ValueError(msg)

        parsed_entry = module.variables.get(entry_value)

        if parsed_entry is None:
            msg = f"entry not found, got {entry_value}"
            raise KeyError(msg)

        return parsed_entry

    def _resolve_list(self, key: str, entry: list) -> FieldEntry[list]:
        return FieldEntry(
            key=key,
            value=ListEntry([self._resolve_entry(key, e) for e in entry]),
        )

    def _resolve_args(self, key: str, args: Any) -> DictEntry[str] | ListEntry:
        if isinstance(args, dict):
            resolved = DictEntry()
            for k, v in args.items():
                value = (
                    FieldEntry(key=key, value=self.kwargs[k])
                    if k in self.kwargs
                    else self._resolve_entry(key, v)
                )

                resolved[k] = value
                self.variables[k] = value

        if isinstance(args, list):
            resolved = ListEntry()
            for v in args:
                resolved.append(self._resolve_entry(key, v))

        return resolved

    def _resolve_dict(self, key: str, entry: dict[str]) -> Entry:
        module_path = entry.get("module")
        symbol_name = entry.get("source")

        if module_path is None or symbol_name is None:
            return FieldEntry(
                key=key,
                value=DictEntry({k: self._resolve_entry(key, v) for k, v in entry.items()}),
            )

        call = entry.get("call", True)
        lazy = entry.get("lazy", False)
        args = entry.get("args", {})

        obj = self._load_symbol_from_module(module_path, symbol_name)

        resolved_args = self._resolve_args(key, args)

        return ModuleEntry(key=key, obj=obj, call=call, lazy=lazy, args=resolved_args)

    def _resolve_entry(self, key: str, entry: Any) -> Entry:
        if isinstance(entry, str):
            return self._resolve_string(key, entry)

        if isinstance(entry, list):
            return self._resolve_list(key, entry)

        if isinstance(entry, dict):
            return self._resolve_dict(key, entry)

        return FieldEntry(key=key, value=entry)

    def parse(self) -> DictEntry[str]:
        res = DictEntry()

        for k in self.config:
            value = self._resolve_entry(k, self.config[k])

            res[k] = value
            self.variables[k] = value

        return res



from pathlib import Path

from kaizo import ConfigParser

VAL = 4

config = f"""
val: {VAL}
use_args:
  module: math
  source: sqrt
  call: true
  args:
    - .val
"""


def test_args_reference(tmp_path: Path) -> None:
    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    assert out["val"] == VAL
    assert out["use_args"] == VAL**0.5



from pathlib import Path

from kaizo import ConfigParser

RESULT = 1

main_py = """
num_call = 0
def fn():
    global num_call

    num_call+=1
    return num_call

def fn2(n):
    return n
"""

config = """
local: main.py
run01:
  module: local
  source: fn2
  call: true
  args:
    n:
        module: local
        source: fn
run02:
    module: local
    source: fn2
    args:
        n: .n
"""


def test_num_call(tmp_path: Path) -> None:
    module = tmp_path / "main.py"
    module.write_text(main_py)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    assert "run02" in out
    assert out["run02"] == RESULT
